"""This module contains functions for parsing the datetime formats used by the CDBS
HTML tables and HST FITS data sets.   It contains a formatting function which will render
a date and time in a sortable string representation (isoformat).
"""
import datetime
import re

from astropy.time import Time

from . import config, exceptions, log

# =======================================================================

def reformat_date(date, sep=" "):
    """Reformat datestring `d` in any recognized format in CRDS standard form.
    Also handles datetime objects and astropy Time."""
    parsed = parse_date(date)
    return format_date(parsed, sep=sep)

def format_date(date, sep=" "):
    """Format a datestring `d` in CRDS standard form.  Also handles datetime
    objects and astropy Time.

    >>> format_date("Mar 21 2001 12:00:00 am")
    '2001-03-21 00:00:00'
    """
    if isinstance(date, str):
        date = parse_date(date)
    return date.isoformat(sep)

T_SEPARATED_DATE_RE = re.compile(r"^\d\d\d\d[-/]\d\d[-/]\d\dT\d\d(:\d\d){1,2}(\.\d{1,6})?$")
ALPHABETICAL_RE = re.compile(r"[A-Za-z]{3,10}")
ASTROPY_TIME_FORMAT = "%Y-%m-%d %H:%M:%S"

def parse_date(date):
    """Parse any date-time string into a datetime object.  Also
    handles datetime objects and astropy Time.

    >>> parse_date("Dec 01 1993 00:00:00 UT")
    datetime.datetime(1993, 12, 1, 0, 0)

    >>> parse_date('Feb 08 2006 01:02AM')
    datetime.datetime(2006, 2, 8, 1, 2)

    >>> parse_date('12/21/1999 05:42:35')
    datetime.datetime(1999, 12, 21, 5, 42, 35)

    >>> parse_date('1999-12-21T05:42:35')
    datetime.datetime(1999, 12, 21, 5, 42, 35)

    >>> parse_numerical_date('12-21-1999 05:42')
    datetime.datetime(1999, 12, 21, 5, 42)

    >>> parse_date("19970114:053714")
    datetime.datetime(1997, 1, 14, 5, 37, 14)

    >>> dtval = parse_date(datetime.datetime.now())
    >>> isinstance(dtval, datetime.datetime)
    True

    >>> parse_date(Time("1999-12-21T05:42:35"))
    datetime.datetime(1999, 12, 21, 5, 42, 35)

    >>> parse_date(Time("1999-12-21T05:42:35.123"))
    datetime.datetime(1999, 12, 21, 5, 42, 35)

    >>> parse_date(Time(2451533.738647963, format="jd", scale="tt"))
    datetime.datetime(1999, 12, 21, 5, 42, 35)

    >>> parse_date("2008-10-15T08:44:44.619 UNDEFINED")
    Traceback (most recent call last):
    ...
    InvalidDatetimeError: One or more required date/time values is UNDEFINED

    """
    if isinstance(date, datetime.datetime):
        date = str(date)

    if isinstance(date, Time):
        date = date.utc.strftime(ASTROPY_TIME_FORMAT)

    if "UNDEFINED" in date:
        raise exceptions.InvalidDatetimeError(
            "One or more required date/time values is UNDEFINED")

    if date.endswith(" UT"):  # Dec 01 1993 00:00:00 UT
        date = date[:-3]

    if T_SEPARATED_DATE_RE.match(date):
        date = date.replace("T", " ")

    if ALPHABETICAL_RE.search(date):
        return parse_alphabetical_date(date)
    else:
        return parse_numerical_date(date)

def now(sep=" "):
    """Returns the timestamp for the current time."""
    return format_date(datetime.datetime.now(), sep)

MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun",
          "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

def month_num(month, initial_date=None):
    """Convert a month name to a month number."""
    try:
        return  MONTHS.index(month[:3].capitalize()) + 1
    except ValueError as exc:
        raise ValueError(
            "Invalid month value " + repr(month) + " in base date " +
            repr(initial_date)) from exc

def parse_alphabetical_date(date):
    """Parse date/time strings with alphabetical months into datetime's.

    >>> parse_alphabetical_date('Feb 08 2006 01:02AM')
    datetime.datetime(2006, 2, 8, 1, 2)

    >>> parse_alphabetical_date('feb 08 2006')
    datetime.datetime(2006, 2, 8, 0, 0)

    >>> parse_alphabetical_date('July 27, 1999 00:00:00')
    datetime.datetime(1999, 7, 27, 0, 0)

    >>> parse_alphabetical_date('JULY 27, 1999 00:00:00')
    datetime.datetime(1999, 7, 27, 0, 0)

    >>> parse_alphabetical_date("03-Jan-2013")
    datetime.datetime(2013, 1, 3, 0, 0)

    >>> format_date("Mar 21 2001 12:00:00 am")
    '2001-03-21 00:00:00'
    """
    initial_date = date
    while date.lower().endswith(" am"):
        date = date.lower().replace(" am", "am")
    while date.lower().endswith(" pm"):
        date = date.lower().replace(" pm", "pm")

    date = date.replace("-", " ")
    date = date.replace("/", " ")

    try:
        month, day, year, time = date.split()    # 'Feb 08 2006 01:02AM'
    except ValueError:
        month, day, year = date.split()          # 'Feb 08 2006'
        time = "00:00AM"

    if day.endswith(","):     #
        day = day[:-1]

    try:
        imonth = month_num(month, initial_date)
        iday = int(day)
        iyear = int(year)
    except Exception:
        day, month = month, day
        imonth = month_num(month, initial_date)
        iday = int(day)
        iyear = int(year)

    ihour, iminute, isecond, imicrosecond = parse_time(time)

    return datetime.datetime(iyear, imonth, iday, ihour, iminute, isecond,
                             imicrosecond)

def parse_time(time):
    """Parse a time string into (hour, minute, second, microsecond),
    including AM/PM.

    >>> parse_time('12:00')
    (12, 0, 0, 0)

    >>> parse_time('01:02AM')
    (1, 2, 0, 0)

    >>> parse_time('01:02PM')
    (13, 2, 0, 0)

    >>> parse_time('13:02PM')
    (13, 2, 0, 0)

    >>> parse_time('13:02 PM')
    (13, 2, 0, 0)

    >>> parse_time(' 13:02 PM ')
    (13, 2, 0, 0)

    >>> parse_time('12:42 AM')
    (0, 42, 0, 0)

    """
    time = time.strip()
    ampm = "unspecified"
    if time[-2:].upper() in ["AM", "PM"]:
        ampm = time[-2:].upper()
        time = time[:-2]
    try:
        hour, minute = time.split(":")
        second = "00"
    except ValueError:
        try:
            hour, minute, second = time.split(":")
        except ValueError:
            hour, minute, second, subsec = time.split(":")
    ihour = int(hour)
    iminute = int(minute)
    second = float(second)
    isecond = int(second)   # int rejects float strs
    imicrosecond = int((second-isecond) * 10**6)
    if ampm == "PM":
        if ihour < 12:
            ihour += 12
    elif ampm == "AM":
        if ihour == 12:
            ihour -= 12
    return ihour, iminute, isecond, imicrosecond

MONTH_DAY_YEAR_RE = re.compile(r"^\d\d/\d\d/\d\d\d\d$")
YEAR_MONTH_DAY_RE = re.compile(r"^\d\d\d\d/\d\d/\d\d$")
NINETEEN_HUNDREDS_RE = re.compile(r"^\d\d/\d\d/9\d$")
TWENTY_FIRST_CENT_RE = re.compile(r"^\d\d/\d\d/[0-3]\d$")
DATE_COLON_TIME_RE = re.compile(r"^\d\d\d\d\d\d\d\d:\d\d\d\d\d\d$")

def parse_numerical_date(dstr):
    """Parse a datetime string with the month expressed as a number in various
    formats.   Return a datetime object.

    >>> parse_numerical_date('12/21/1999 05:42')
    datetime.datetime(1999, 12, 21, 5, 42)

    >>> parse_numerical_date('12/21/1999 05:42:35')
    datetime.datetime(1999, 12, 21, 5, 42, 35)

    >>> parse_numerical_date('1999-12-21 05:42:00')
    datetime.datetime(1999, 12, 21, 5, 42)

    >>> parse_numerical_date('12-21-1999 05:42:00')
    datetime.datetime(1999, 12, 21, 5, 42)

    >>> parse_numerical_date('21/12/99 05:42:00 PM')
    datetime.datetime(1999, 12, 21, 17, 42)

    >>> parse_numerical_date('21/12/01 05:42:00')
    datetime.datetime(2001, 12, 21, 5, 42)

    >>> parse_numerical_date('21/12/15 05:42:00')
    datetime.datetime(2015, 12, 21, 5, 42)

    >>> parse_date("19970114:053714")
    datetime.datetime(1997, 1, 14, 5, 37, 14)
    """
    dstr = dstr.replace("-","/")
    parts = dstr.split()
    date = parts[0]
    if len(parts) == 1:
        time = "00:00:00"
    else:
        time = " ".join(parts[1:])

    ihour, iminute, isecond, imicrosecond = parse_time(time)

    if MONTH_DAY_YEAR_RE.match(date):
        month, day, year = date.split("/")
    elif YEAR_MONTH_DAY_RE.match(date):
        year, month, day = date.split("/")
    elif NINETEEN_HUNDREDS_RE.match(date):
        day, month, year = date.split("/")
        year = "19" + year
    elif TWENTY_FIRST_CENT_RE.match(date):
        day, month, year = date.split("/")
        year = "20" + year
    elif DATE_COLON_TIME_RE.match(date):
        year = date[:4]
        month = date[4:6]
        day = date[6:8]
        ihour = int(date[9:11])
        iminute = int(date[11:13])
        isecond = int(date[13:15])
        imicrosecond = 0
    else:
        raise ValueError("Unknown numerical date format: " + repr(dstr))

    imonth, iday, iyear, = int(month), int(day), int(year)

    return datetime.datetime(iyear, imonth, iday, ihour, iminute, isecond,
                             imicrosecond)

# ============================================================================

class DateParser:
    """Abstract baseclass for defining date parsers."""

    format = re.compile("^$")
    should_be = "DATE FORMAT NOT DEFINED"

    @classmethod
    def _get_date_dict(cls, match):
        raise NotImplementedError("Data Parser is an abstract class.")

    @classmethod
    def get_datetime(cls, datestr):
        match = cls.format.match(datestr)
        if not match:
            raise ValueError(
                "Invalid " + repr(cls.__name__) + " format " + repr(datestr) +
                " should be " + repr(cls.should_be)
            )
        return datetime.datetime(**cls._get_date_dict(match))

class Slashdate(DateParser):
    """
    >>> Slashdate.get_datetime("25 / 12 / 2014")
    datetime.datetime(2014, 12, 25, 0, 0)

    >>> Slashdate.get_datetime(r"25 / 12 x 2014")
    Traceback (most recent call last):
    ...
    ValueError: Invalid 'Slashdate' format '25 / 12 x 2014'
    """
    format = re.compile(r"(?P<day>\d\d)\s*/\s*(?P<month>\d\d)\s*/\s*(?P<year>\d\d\d\d)")
    should_be = "DD/MM/YYYY"

    @classmethod
    def _get_date_dict(cls, match):
        return dict(month=int(match.group("month")),
                    day=int(match.group("day")),
                    year=int(match.group("year")))

class Dashdate(DateParser):
    """Originally Slashdate supported HST PEDIGREE dates, later extended to JWST PEDIGREE dates.

    >>> Dashdate.get_datetime("2014-12-25")
    datetime.datetime(2014, 12, 25, 0, 0)

    >>> Dashdate.get_datetime(r"25 / 12 x 2014")
    Traceback (most recent call last):
    ...
    ValueError: Invalid 'Ddate' format '25 / 12 x 2014'
    """
    format = re.compile(r"(?P<year>\d\d\d\d)\-(?P<month>\d\d)\-(?P<day>\d\d)")
    should_be = "YYYY-MM-DD"

    @classmethod
    def _get_date_dict(cls, match):
        return dict(month=int(match.group("month")),
                    day=int(match.group("day")),
                    year=int(match.group("year")))

class Sybdate(DateParser):
    """
    >>> Sybdate.get_datetime("Mar 21 2001")
    datetime.datetime(2001, 3, 21, 0, 0)

    >>> Sybdate.get_datetime("Mar 21 2001 12:00:00")
    datetime.datetime(2001, 3, 21, 12, 0)

    >>> Sybdate.get_datetime("Mar 21 2001 12:00:00 am")
    datetime.datetime(2001, 3, 21, 0, 0)

    >>> Sybdate.get_datetime("Mar 21 2001 12:00:00 PM")
    datetime.datetime(2001, 3, 21, 12, 0)

    >>> Sybdate.get_datetime("Mar 21 2001 01:00:00 PM")
    datetime.datetime(2001, 3, 21, 13, 0)

    >>> Sybdate.get_datetime("Mxx 21 2001 01:00:00 PM")
    Traceback (most recent call last):
    ...
    ValueError: Invalid month value 'Mxx'
    """
    format = re.compile(
        r"(?P<month>[A-Za-z]{1,10})\s+" + \
            r"(?P<day>\d{1,2}),?\s+" + \
            r"(?P<year>\d{1,4})" + \
            r"(\s+(?P<hour>\d{1,2}):" + \
            r"(?P<minute>\d{1,2}):" + \
            r"(?P<second>\d{1,2})\s*" + \
            r"(?P<meridian>am|pm|AM|PM)?" + \
            r")?")

    should_be ="Mar 21 2001 12:00:00 am"

    @classmethod
    def _get_date_dict(cls, match):
        items = match.groupdict()
        try:
            items["month"] = month_num(match.group("month"))
        except IndexError as exc:
            raise ValueError(
                "Illegal month", repr(match.group("month"))) from exc
        if items["meridian"]:
            hour = int(items["hour"])
            if items["meridian"].lower() == "pm":
                if hour < 12:
                    hour += 12
            else:
                if hour == 12:
                    hour -= 12
            items["hour"] = str(hour)
            del items["meridian"]
        return dict([(name, int(x)) for (name, x) in items.items() if x])

class Jwstdate(DateParser):
    """
    >>> Jwstdate.get_datetime("2001-03-21T00:00:00")
    datetime.datetime(2001, 3, 21, 0, 0)

    >>> Jwstdate.get_datetime("2001-03-21")
    datetime.datetime(2001, 3, 21, 0, 0)

    >>> Jwstdate.get_datetime("2001-03-21 12:00:00")
    Traceback (most recent call last):
    ...
    ValueError: Invalid 'Jwstdate' format '2001-03-21 12:00:00'
    """
    format = re.compile(
        r"^(?P<year>\d\d\d\d)\-" + \
            r"(?P<month>\d\d)\-" + \
            r"(?P<day>\d\d)(T" + \
            r"(?P<hour>\d\d):" + \
            r"(?P<minute>\d\d):" + \
            r"(?P<second>\d\d))?$")

    should_be = "2018-12-22T00:00:00"

    @classmethod
    def _get_date_dict(cls, match):
        items = match.groupdict()
        return dict([(name, int(x)) for (name, x) in items.items() if x])

class Anydate(DateParser):
    """
    Historically Anydate was an HST format limited to Sybdate and Slashdate:

    >>> Anydate.get_datetime("25 / 12 / 2014")
    datetime.datetime(2014, 12, 25, 0, 0)

    >>> Anydate.get_datetime("Mar 21 2001 12:00:00")
    datetime.datetime(2001, 3, 21, 12, 0)

    Consequently, the CRDS (and JWST FITS) internal time format is not an Anydate:

    >>> Anydate.get_datetime("2001-03-21T00:00:00")
    Traceback (most recent call last):
    ...
    ValueError: Invalid Anydate format '2001-03-21T00:00:00'

    This is controversial but conservatism suggests limiting the format to historical possibilities in
    case there is any downstream software incapable of handling the newer CRDS/FITS format.
    """

    @classmethod
    def get_datetime(cls, datestr):
        try:
            return Slashdate.get_datetime(datestr)
        except ValueError:
            pass
        try:
            return Sybdate.get_datetime(datestr)
        except ValueError as exc:
            raise ValueError("Invalid Anydate format " + repr(datestr)) from exc


#class Shortdate(DateParser):
#    pass
#
#class DashDate(DateParser):
#    pass
#
#class CdbsDate(DateParser):
#    pass

# ============================================================================

DATETIME_RE_STR = r"^(\d\d\d\d\-\d\d\-\d\d(\s+|T)\d\d:\d\d:\d\d)$"
DATETIME_RE = re.compile(DATETIME_RE_STR)
DATE_RE_STR = r"^\d\d\d\d\-\d\d\-\d\d$"
TIME_RE_STR = r"^\d\d:\d\d:\d\d$"

def is_datetime(datetime_str):
    """Raise an assertion error if `datetime_str` doesn't look like a CRDS date.
    Otherwise return `datetime_str`.   This is used to match the composite
    datetime string used by the UseAfter selector.

    >>> is_datetime('2001-03-21T00:00:00')
    '2001-03-21T00:00:00'

    >>> is_datetime('2001-03-21 00:00:00')
    '2001-03-21 00:00:00'

    >>> is_datetime('2001-03-99 00:00:00')
    Traceback (most recent call last):
    ...
    CrdsError: day is out of range for month

    >>> is_datetime('2001-03-21X00:00:00')
    Traceback (most recent call last):
    ...
    AssertionError: Invalid date/time.  Should be YYYY-MM-DD HH:MM:SS
    """
    assert DATETIME_RE.match(datetime_str), \
        "Invalid date/time.  Should be YYYY-MM-DD HH:MM:SS"
    try:
        parse_date(datetime_str)
    except ValueError as exc:
        raise exceptions.CrdsError(str(exc)) from exc
    return datetime_str


# ============================================================================
def reformat_useafter(filename, header):
    """Reformat a USEAFTER date in a standard CRDS form which can be split into
    DATE-OBS and TIME-OBS.   Honor the ALLOW_BAD_USEAFTER to provide a safe default
    for early junk USEAFTER values;  1900-01-01T00:00:00.
    """
    useafter = str(header.get("USEAFTER", header.get("META.USEAFTER", header.get("ROMAN.META.USEAFTER", "UNDEFINED"))))
    try:
        return reformat_date(useafter)
    except Exception as exc:
        if config.ALLOW_BAD_USEAFTER:
            log.warning("Can't parse USEAFTER =", repr(useafter),
                        "for", repr(filename), "faking as '1900-01-01T00:00:00'")
            return reformat_date("1900-01-01T00:00:00")
        else:
            raise exceptions.InvalidUseAfterFormat(
                "Bad USEAFTER time format =", repr(useafter)) from exc

def get_slash_date(datestr):
    """Return the datetime object corresponding to `datestr` which must
    be in DD/MM/YYYY format.
    """
    return Slashdate.get_datetime(datestr)

def get_dash_date(datestr):
    """Return the datetime object corresponding to `datestr` which must
    be in YYYY-MM-DD format.
    """
    return Dashdate.get_datetime(datestr)

# ============================================================================

def test():
    """Run module doctests."""
    import doctest
    from crds.core import timestamp
    return doctest.testmod(timestamp)

if __name__ == "__main__":
    print(test())



import numpy as np
from astropy.io import fits

from crds.misc import synphot
from crds.tests import test_config

from nose.tools import assert_true


class TestSynphotLookupGenerator(test_config.CRDSTestCase):
    def test_throughput_lookup_generation(self):
        generator = synphot.SynphotLookupGenerator("hst_0779.pmap")

        original_hdul = generator.generate("tmctab", [])
        original_tmc = original_hdul[-1].data
        original_row = original_tmc[original_tmc["COMPNAME"] == "wfpc_pol60"][0]
        assert_true(original_row["FILENAME"] == "crwfpccomp$wfpc_pol60_003_syn.fits")

        new_hdul = generator.generate("tmctab", ["data/hst_synphot_new_syn.fits"])
        new_tmc = new_hdul[-1].data
        new_row = new_tmc[new_tmc["COMPNAME"] == "wfpc_pol60"][0]
        assert_true(new_row["FILENAME"] == "crwfpccomp$hst_synphot_new_syn.fits")

    def test_thermal_lookup_generation(self):
        generator = synphot.SynphotLookupGenerator("hst_0779.pmap")

        original_hdul = generator.generate("tmttab", [])
        original_tmt = original_hdul[-1].data
        original_row = original_tmt[original_tmt["COMPNAME"] == "wfc3_ir_f167n"][0]
        assert_true(original_row["FILENAME"] == "crwfc3comp$wfc3_ir_f167n_002_th.fits")

        new_hdul = generator.generate("tmttab", ["data/hst_synphot_new_th.fits"])
        new_tmt = new_hdul[-1].data
        new_row = new_tmt[new_tmt["COMPNAME"] == "wfc3_ir_f167n"][0]
        assert_true(new_row["FILENAME"] == "crwfc3comp$hst_synphot_new_th.fits")



import doctest

"""
usage: /Users/jmiller/work/normal/lib/python2.7/site-packages/crds/diff.py
       [-h] [-P] [-T] [-K] [-v] [--verbosity VERBOSITY] [-V] [-J] [-H]
       [--profile PROFILE] [--pdb]
       old_file new_file

Difference CRDS mapping or reference files.

positional arguments:
  old_file              Prior file of difference.
  new_file              New file of difference.

optional arguments:
  -h, --help            show this help message and exit
  -P, --primitive-diffs
                        Fitsdiff replaced reference files when diffing mappings.
  -T, --mapping-text-diffs
                        In addition to CRDS mapping logical differences,  run UNIX context diff for mappings.
  -K, --check-diffs     Issue warnings about new rules, deletions, or reversions.
  -v, --verbose         Set log verbosity to True,  nominal debug level.
  --verbosity VERBOSITY
                        Set log verbosity to a specific level: 0..100.
  -V, --version         Print the software version and exit.
  -J, --jwst            Force observatory to JWST for determining header conventions.
  -H, --hst             Force observatory to HST for determining header conventions.
  --profile PROFILE     Output profile stats to the specified file.
  --pdb                 Run under pdb.

Reference files are nominally differenced using FITS-diff or diff.

Mapping files are differenced using CRDS machinery to recursively compare too mappings and
their sub-mappings.

Differencing two mappings will find all the logical differences between the two contexts
and any nested mappings.

By specifying --mapping-text-diffs,  UNIX diff will be run on mapping files in addition to
CRDS logical diffs.

By specifying --primitive-diffs,  FITS diff will be run on all references which are replaced
in the logical differences between two mappings.

For example:

    % crds diff hst_0001.pmap  hst_0005.pmap  --mapping-text-diffs --primitive-diffs

Will recursively produce logical, textual, and FITS diffs for all changes between the two contexts.

    NOTE: mapping logical differences (the default) to not compare CRDS mapping headers.

----------
TEST CASES
----------
"""

from crds.core import log
from crds import tests
from crds.tests import test_config

from crds.diff import DiffScript

def dt_diff_pmap_diffs():
    """
    Compute diffs for two .pmap's:

    >>> old_state = test_config.setup()
    >>> DiffScript("crds.diff data/hst_0001.pmap data/hst_0002.pmap")()
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00'), 'replaced data/hst_acs_biasfile_0001.fits with data/hst_acs_biasfile_0002.fits')
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'deleted Match rule for m991609tj_bia.fits')
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'deleted Match rule for q9e1206kj_bia.fits')
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53'), 'deleted Match rule for q9e12071j_bia.fits')
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'added Match rule for m991609tj_bia.fits')
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'added Match rule for q9e1206kj_bia.fits')
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54'), 'added Match rule for q9e12071j_bia.fits')
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('biasfile',), 'replaced data/hst_acs_biasfile_0001.rmap with data/hst_acs_biasfile_0002.rmap')
    (('data/hst_0001.pmap', 'data/hst_0002.pmap'), ('acs',), 'replaced data/hst_acs_0001.imap with data/hst_acs_0002.imap')
    1
    >>> test_config.cleanup(old_state)
    """

def dt_diff_imap_diffs():
    """
    Compute diffs for two .imap's:

    >>> old_state = test_config.setup()
    >>> DiffScript("crds.diff data/hst_acs_0001.imap data/hst_acs_0002.imap")()
    (('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00'), 'replaced data/hst_acs_biasfile_0001.fits with data/hst_acs_biasfile_0002.fits')
    (('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'deleted Match rule for m991609tj_bia.fits')
    (('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'deleted Match rule for q9e1206kj_bia.fits')
    (('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53'), 'deleted Match rule for q9e12071j_bia.fits')
    (('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'added Match rule for m991609tj_bia.fits')
    (('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'added Match rule for q9e1206kj_bia.fits')
    (('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54'), 'added Match rule for q9e12071j_bia.fits')
    (('data/hst_acs_0001.imap', 'data/hst_acs_0002.imap'), ('biasfile',), 'replaced data/hst_acs_biasfile_0001.rmap with data/hst_acs_biasfile_0002.rmap')
    1
    >>> test_config.cleanup(old_state)
    """

def dt_diff_rmap_diffs():
    """
    Compute diffs for two .rmap's:

    >>> old_state = test_config.setup()
    >>> DiffScript("crds.diff data/hst_acs_biasfile_0001.rmap data/hst_acs_biasfile_0002.rmap --include-header-diffs")()
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00'), 'replaced data/hst_acs_biasfile_0001.fits with data/hst_acs_biasfile_0002.fits')
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'deleted Match rule for m991609tj_bia.fits')
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'deleted Match rule for q9e1206kj_bia.fits')
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53'), 'deleted Match rule for q9e12071j_bia.fits')
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'added Match rule for m991609tj_bia.fits')
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'added Match rule for q9e1206kj_bia.fits')
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54'), 'added Match rule for q9e12071j_bia.fits')
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), 'deleted header \\'rmap_relevance\\' = \\'((DETECTOR != "SBC") and (BIASCORR != "OMIT"))\\'')
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), "header added 'extra_info' = 'some other piece of information.'")
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), "header replaced 'extra_keys' = ('XCORNER', 'YCORNER', 'CCDCHIP') with ('ZCORNER', 'YCORNER', 'CCDCHIP')")
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), "header replaced 'reffile_required' = 'yes' with 'no'")
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), "header replaced 'sha1sum' = '90a43965be5d044f1e8fcf5f141b3f64b763ca89' with '32df61d8f1cd4d398d84d05e1706b5565712d87d'")
    1
    >>> test_config.cleanup(old_state)
    """

def dt_diff_fits_diff():
    """
    Compute diffs for two .fits's:

    >>> old_state = test_config.setup()

    >>> DiffScript("crds.diff data/hst_acs_biasfile_0001.fits data/hst_acs_biasfile_0002.fits")() # doctest: +ELLIPSIS
    <BLANKLINE>
     fitsdiff: ...
     a: data/hst_acs_biasfile_0001.fits
     b: data/hst_acs_biasfile_0002.fits
     Maximum number of different data values to be reported: 10
     Relative tolerance: 0.0, Absolute tolerance: 0.0
    <BLANKLINE>
    Primary HDU:
    <BLANKLINE>
       Headers contain differences:
         Extra keyword 'ADD_1'  in a: 'added to hst_acs_biasfile_0001.fits'
         Extra keyword 'ADD_2'  in b: 'added to hst_acs_biasfile_0002.fits'
         Keyword DIFF_12  has different values:
            a> value in 1
             ?          ^
            b> value in 2
             ?          ^
    <BLANKLINE>
    <BLANKLINE>
    1

    >>> test_config.cleanup(old_state)
    """

def dt_diff_asdf():
    """
    Compute diffs for two .asdf's:

    >>> old_state = test_config.setup()

    >>> DiffScript("crds.diff data/jwst_nircam_specwcs_0010.asdf data/jwst_nircam_specwcs_0011.asdf")() # doctest: +ELLIPSIS
            ndarrays differ by contents
            ndarrays differ by contents
            ndarrays differ by contents
            ndarrays differ by contents
    tree:
      history:
        -
          description:
    >       Created from NIRCAM_modA_R.conf
    <       Created from NIRCAM_modA_C.conf
          time:
    >       2017-09-08 16:57:27.004949
    <       2017-09-08 16:57:26.927451
            ndarrays differ by contents
            ndarrays differ by contents
            ndarrays differ by contents
            ndarrays differ by contents
      meta:
        date:
    >     2017-09-08T12:57:27.006
    <     2017-09-08T12:57:26.928
        description:
    >     GRISMR dispersion models
    <     GRISMC dispersion models
        filename:
    >     NIRCAM_modA_R.asdf
    <     NIRCAM_modA_C.asdf
        instrument:
          pupil:
    >       GRISMR
    <       GRISMC
    1

     >>> DiffScript("crds.diff data/jwst_nircam_specwcs_0010.asdf data/jwst_nircam_specwcs_0010.asdf")() # doctest: +ELLIPSIS
     0

    >>> test_config.cleanup(old_state)
    """

def dt_diff_rmap_primitive_diffs():
    """
    Compute primitive diffs for two .rmap's:

    >>> old_state = test_config.setup()

    >>> DiffScript("crds.diff data/hst_acs_biasfile_0001.rmap data/hst_acs_biasfile_0002.rmap --primitive-diffs")()  #doctest: +ELLIPSIS
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00'), 'replaced data/hst_acs_biasfile_0001.fits with data/hst_acs_biasfile_0002.fits')
    <BLANKLINE>
     fitsdiff: ...
     a: data/hst_acs_biasfile_0001.fits
     b: data/hst_acs_biasfile_0002.fits
     Maximum number of different data values to be reported: 10
     Relative tolerance: 0.0, Absolute tolerance: 0.0
    <BLANKLINE>
    Primary HDU:
    <BLANKLINE>
       Headers contain differences:
         Extra keyword 'ADD_1'  in a: 'added to hst_acs_biasfile_0001.fits'
         Extra keyword 'ADD_2'  in b: 'added to hst_acs_biasfile_0002.fits'
         Keyword DIFF_12  has different values:
            a> value in 1
             ?          ^
            b> value in 2
             ?          ^
    <BLANKLINE>
    <BLANKLINE>
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'deleted Match rule for m991609tj_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'deleted Match rule for q9e1206kj_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53'), 'deleted Match rule for q9e12071j_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'added Match rule for m991609tj_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'added Match rule for q9e1206kj_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54'), 'added Match rule for q9e12071j_bia.fits')
    1

    >>> test_config.cleanup(old_state)
    """

def dt_diff_file_reversions():
    """
    Compute diffs checking for reversions: (invert file order to simulate reverse filename progression)

    >>> old_state = test_config.setup()

    >>> DiffScript("crds.diff data/hst_0002.pmap data/hst_0001.pmap --check-diffs")()
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00'), 'replaced data/hst_acs_biasfile_0002.fits with data/hst_acs_biasfile_0001.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'added Match rule for m991609tj_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'added Match rule for q9e1206kj_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53'), 'added Match rule for q9e12071j_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'deleted Match rule for m991609tj_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'deleted Match rule for q9e1206kj_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54'), 'deleted Match rule for q9e12071j_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('biasfile',), 'replaced data/hst_acs_biasfile_0002.rmap with data/hst_acs_biasfile_0001.rmap')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('acs',), 'replaced data/hst_acs_0002.imap with data/hst_acs_0001.imap')
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00')) added Match rule for 'm991609tj_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35')) added Match rule for 'q9e1206kj_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53')) added Match rule for 'q9e12071j_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00')) deleted Match rule for 'm991609tj_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35')) deleted Match rule for 'q9e1206kj_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54')) deleted Match rule for 'q9e12071j_bia.fits'
    CRDS - WARNING - Reversion at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00')) replaced 'data/hst_acs_biasfile_0002.fits' with 'data/hst_acs_biasfile_0001.fits'
    CRDS - WARNING - Reversion at ('data/hst_acs_0001.imap', ('biasfile',)) replaced 'data/hst_acs_biasfile_0002.rmap' with 'data/hst_acs_biasfile_0001.rmap'
    CRDS - WARNING - Reversion at ('data/hst_0001.pmap', ('acs',)) replaced 'data/hst_acs_0002.imap' with 'data/hst_acs_0001.imap'
    2

    >>> test_config.cleanup(old_state)
    """

def dt_diff_row_change():
    """
    Row change

    >>> old_state = test_config.setup()

    >>> DiffScript("crds.diff data/test-source.fits data/test-change-row1-valueLeft.fits")()  #doctest: +ELLIPSIS
    <BLANKLINE>
     fitsdiff: ...
     a: data/test-source.fits
     b: data/test-change-row1-valueLeft.fits
     Maximum number of different data values to be reported: 10
     Relative tolerance: 0.0, Absolute tolerance: 0.0
    <BLANKLINE>
    Extension HDU 1:
    <BLANKLINE>
       Data contains differences:
         Column valueLeft data differs in row 1:
            a> 5748
            b> -1
         1 different table data element(s) found (2.22% different).
    <BLANKLINE>
     Row differences for HDU extension #1
    <BLANKLINE>
        Summary:
            a rows 1-1 differ from b rows 1-1
    <BLANKLINE>
        Row difference, unified diff format:
            --- Table A
    <BLANKLINE>
            +++ Table B
    <BLANKLINE>
            @@ -1,5 +1,5 @@
    <BLANKLINE>
             'yes', 'yes', 2988, -2779.03..., 'coquille'
            -'yes', 'no', 5748, 6357.97..., 'ferly'
            +'yes', 'no', -1, 6357.97..., 'ferly'
             'yes', 'maybe', 9735, -9132.53..., 'misreliance'
             'no', 'yes', 425, -2689.26..., 'ogeed'
             'no', 'no', 8989, 9870.02..., 'readmittance'
    <BLANKLINE>
    1

    >>> test_config.cleanup(old_state)
    """

def dt_diff_rmap_primitive_diffs():
    """
    Compute primitive diffs for two .rmap's:

    >>> old_state = test_config.setup()

    >>> DiffScript("crds.diff data/hst_acs_biasfile_0001.rmap data/hst_acs_biasfile_0002.rmap --primitive-diffs")()  #doctest: +ELLIPSIS
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00'), 'replaced data/hst_acs_biasfile_0001.fits with data/hst_acs_biasfile_0002.fits')
    <BLANKLINE>
     fitsdiff: ...
     a: data/hst_acs_biasfile_0001.fits
     b: data/hst_acs_biasfile_0002.fits
     Maximum number of different data values to be reported: 10
     Relative tolerance: 0.0, Absolute tolerance: 0.0
    <BLANKLINE>
    Primary HDU:
    <BLANKLINE>
       Headers contain differences:
         Extra keyword 'ADD_1'  in a: 'added to hst_acs_biasfile_0001.fits'
         Extra keyword 'ADD_2'  in b: 'added to hst_acs_biasfile_0002.fits'
         Keyword DIFF_12  has different values:
            a> value in 1
             ?          ^
            b> value in 2
             ?          ^
    <BLANKLINE>
    <BLANKLINE>
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'deleted Match rule for m991609tj_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'deleted Match rule for q9e1206kj_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53'), 'deleted Match rule for q9e12071j_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'added Match rule for m991609tj_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'added Match rule for q9e1206kj_bia.fits')
    ================================================================================
    (('data/hst_acs_biasfile_0001.rmap', 'data/hst_acs_biasfile_0002.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54'), 'added Match rule for q9e12071j_bia.fits')
    1

    >>> test_config.cleanup(old_state)
    """

def dt_diff_file_reversions():
    """
    Compute diffs checking for reversions: (invert file order to simulate reverse filename progression)

    >>> old_state = test_config.setup()

    >>> DiffScript("crds.diff data/hst_0002.pmap data/hst_0001.pmap --check-diffs")()
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00'), 'replaced data/hst_acs_biasfile_0002.fits with data/hst_acs_biasfile_0001.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'added Match rule for m991609tj_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'added Match rule for q9e1206kj_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53'), 'added Match rule for q9e12071j_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00'), 'deleted Match rule for m991609tj_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35'), 'deleted Match rule for q9e1206kj_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('data/hst_acs_biasfile_0002.rmap', 'data/hst_acs_biasfile_0001.rmap'), ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54'), 'deleted Match rule for q9e12071j_bia.fits')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('data/hst_acs_0002.imap', 'data/hst_acs_0001.imap'), ('biasfile',), 'replaced data/hst_acs_biasfile_0002.rmap with data/hst_acs_biasfile_0001.rmap')
    (('data/hst_0002.pmap', 'data/hst_0001.pmap'), ('acs',), 'replaced data/hst_acs_0002.imap with data/hst_acs_0001.imap')
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00')) added Match rule for 'm991609tj_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35')) added Match rule for 'q9e1206kj_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '4.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:42:53')) added Match rule for 'q9e12071j_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('1992-01-01', '00:00:00')) deleted Match rule for 'm991609tj_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-04', '11:32:35')) deleted Match rule for 'q9e1206kj_bia.fits'
    CRDS - WARNING - Rule change at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '5.0', '*', '1062', '1044', '19.0', '20.0', 'N/A', 'N/A', 'N/A'), ('2006-07-15', '04:43:54')) deleted Match rule for 'q9e12071j_bia.fits'
    CRDS - WARNING - Reversion at ('data/hst_acs_biasfile_0001.rmap', ('HRC', 'A', '1.0', '*', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A'), ('1992-01-02', '00:00:00')) replaced 'data/hst_acs_biasfile_0002.fits' with 'data/hst_acs_biasfile_0001.fits'
    CRDS - WARNING - Reversion at ('data/hst_acs_0001.imap', ('biasfile',)) replaced 'data/hst_acs_biasfile_0002.rmap' with 'data/hst_acs_biasfile_0001.rmap'
    CRDS - WARNING - Reversion at ('data/hst_0001.pmap', ('acs',)) replaced 'data/hst_acs_0002.imap' with 'data/hst_acs_0001.imap'
    2
    >>> test_config.cleanup(old_state)
    """

def dt_diff_row_change():
    """
    Row change

    >>> old_state = test_config.setup()

    >>> DiffScript("crds.diff data/test-source.fits data/test-change-row1-valueLeft.fits")()  #doctest: +ELLIPSIS
    <BLANKLINE>
     fitsdiff: ...
     a: data/test-source.fits
     b: data/test-change-row1-valueLeft.fits
     Maximum number of different data values to be reported: 10
     Relative tolerance: 0.0, Absolute tolerance: 0.0
    <BLANKLINE>
    Extension HDU 1:
    <BLANKLINE>
       Data contains differences:
         Column valueLeft data differs in row 1:
            a> 5748
            b> -1
         1 different table data element(s) found (2.22% different).
    <BLANKLINE>
     Row differences for HDU extension #1
    <BLANKLINE>
        Summary:
            a rows 1-1 differ from b rows 1-1
    <BLANKLINE>
        Row difference, unified diff format:
            --- Table A
    <BLANKLINE>
            +++ Table B
    <BLANKLINE>
            @@ -1,5 +1,5 @@
    <BLANKLINE>
             'yes', 'yes', 2988, -2779.03..., 'coquille'
            -'yes', 'no', 5748, 6357.97..., 'ferly'
            +'yes', 'no', -1, 6357.97..., 'ferly'
             'yes', 'maybe', 9735, -9132.53..., 'misreliance'
             'no', 'yes', 425, -2689.26..., 'ogeed'
             'no', 'no', 8989, 9870.02..., 'readmittance'
    <BLANKLINE>
    1
    >>> test_config.cleanup(old_state)
    """

def dt_diff_print_affected_modes():
    """
    >>> old_state = test_config.setup()
    >>> DiffScript("crds.diff data/hst_cos_deadtab.rmap data/hst_cos_deadtab_9998.rmap --print-affected-modes")()
    INSTRUMENT='COS' REFTYPE='DEADTAB' DETECTOR='FUV' DIFF_COUNT='1'
    INSTRUMENT='COS' REFTYPE='DEADTAB' DETECTOR='NUV' DIFF_COUNT='1'
    1
    >>> test_config.cleanup(old_state)
    """

def dt_diff_print_all_new_files():
    """
    >>> old_state = test_config.setup(cache=test_config.CRDS_TESTING_CACHE)
    >>> DiffScript("crds.diff data/hst_0001.pmap data/hst_0008.pmap --print-all-new-files --sync-files --include-header-diffs --hide-boring")()
    CRDS - INFO - 0 errors
    CRDS - INFO - 0 warnings
    CRDS - INFO - 0 infos
    hst_0002.pmap
    hst_0003.pmap
    hst_0004.pmap
    hst_0005.pmap
    hst_0006.pmap
    hst_0007.pmap
    hst_0008.pmap
    hst_acs.imap acs
    hst_acs_0002.imap acs
    hst_acs_biasfile.rmap acs biasfile
    hst_acs_biasfile_0002.rmap acs biasfile
    hst_cos_0001.imap cos
    hst_cos_0002.imap cos
    hst_cos_flatfile_0002.rmap cos flatfile
    hst_cos_flatfile_0003.rmap cos flatfile
    hst_cos_fluxtab.rmap cos fluxtab
    hst_cos_xtractab.rmap cos xtractab
    hst_stis_0001.imap stis
    hst_stis_0002.imap stis
    hst_stis_biasfile_0001.rmap stis biasfile
    hst_stis_biasfile_0002.rmap stis biasfile
    hst_stis_darkfile_0001.rmap stis darkfile
    hst_stis_darkfile_0002.rmap stis darkfile
    hst_wfc3_0001.imap wfc3
    hst_wfc3_0002.imap wfc3
    hst_wfc3_0003.imap wfc3
    hst_wfc3_darkfile_0001.rmap wfc3 darkfile
    hst_wfc3_darkfile_0004.rmap wfc3 darkfile
    hst_wfc3_flshfile_0001.rmap wfc3 flshfile
    data/hst_acs_biasfile_0002.fits acs biasfile
    s7g17006l_1dx.fits cos xtractab
    s7g17007l_1dx.fits cos xtractab
    s7g1700kl_phot.fits cos fluxtab
    s7g1700nl_1dx.fits cos xtractab
    s7g1700ol_1dx.fits cos xtractab
    s7g1700tl_flat.fits cos flatfile
    s7g17011l_phot.fits cos fluxtab
    t9h1220sl_phot.fits cos fluxtab
    u8k1433ql_phot.fits cos fluxtab
    w5g1439sl_1dx.fits cos xtractab
    x1v17414l_1dx.fits cos xtractab
    x1v17416l_phot.fits cos fluxtab
    x5v1944hl_flat.fits cos flatfile
    x6q17586l_1dx.fits cos xtractab
    x6q17587l_phot.fits cos fluxtab
    x7h1457mi_drk.fits wfc3 darkfile
    x7h1457ni_drk.fits wfc3 darkfile
    x7h1457oi_drk.fits wfc3 darkfile
    x7h1457pi_drk.fits wfc3 darkfile
    x7v2004ri_drk.fits wfc3 darkfile
    x7v2004si_drk.fits wfc3 darkfile
    x7v2004ti_drk.fits wfc3 darkfile
    x8c16266o_bia.fits stis biasfile
    x8c16267o_bia.fits stis biasfile
    x8c16268o_bia.fits stis biasfile
    x8c16269o_bia.fits stis biasfile
    x8c1626ao_bia.fits stis biasfile
    x8c1626bo_bia.fits stis biasfile
    x8c1626co_drk.fits stis darkfile
    x8c1626do_drk.fits stis darkfile
    x8c1626eo_drk.fits stis darkfile
    x8c1626fo_drk.fits stis darkfile
    x8k1551hi_drk.fits wfc3 darkfile
    x8k1551ii_drk.fits wfc3 darkfile
    x8k1551ji_drk.fits wfc3 darkfile
    x8k1551ki_drk.fits wfc3 darkfile
    x8k1551li_drk.fits wfc3 darkfile
    x9618368i_drk.fits wfc3 darkfile
    x9618369i_drk.fits wfc3 darkfile
    x961836ai_drk.fits wfc3 darkfile
    x961836bi_drk.fits wfc3 darkfile
    x9c1801qo_bia.fits stis biasfile
    x9c1801ro_bia.fits stis biasfile
    x9c1801so_bia.fits stis biasfile
    x9c1801to_bia.fits stis biasfile
    x9c18020o_bia.fits stis biasfile
    x9c18021o_drk.fits stis darkfile
    x9c18022o_drk.fits stis darkfile
    x9c18023o_drk.fits stis darkfile
    1

    >>> test_config.cleanup(old_state)
    """

def dt_diff_print_new_files():
    """
    >>> old_state = test_config.setup()
    >>> DiffScript("crds.diff data/hst_0001.pmap data/hst_0002.pmap --print-new-files")()
    hst_0002.pmap
    hst_acs_0002.imap
    hst_acs_biasfile_0002.rmap
    data/hst_acs_biasfile_0002.fits
    1
    >>> test_config.cleanup(old_state)
    """

def dt_diff_print_affected_types():
    """
    >>> old_state = test_config.setup()
    >>> DiffScript("crds.diff data/hst_cos_deadtab.rmap data/hst_cos_deadtab_9998.rmap --print-affected-types")()
    cos        deadtab
    1
    >>> test_config.cleanup(old_state)
    """

def dt_diff_print_affected_instruments():
    """
    >>> old_state = test_config.setup()
    >>> DiffScript("crds.diff data/hst_cos_deadtab.rmap data/hst_cos_deadtab_9998.rmap --print-affected-instruments")()
    cos
    1
    >>> test_config.cleanup(old_state)
    """

def dt_diff_recurse_added_deleted_na():
    """
    For this test,  checking recursive terminal adds/deletes and N/A + OMIT at all levels:

    Changed hst_wfpc2.imap to N/A
    Replaced hst_cos.imap with data/hst_cos.imap results in recursive COS changes
    Deleted hst_cos_fluxtab.rmap,  results in deleted terminals, rmap
    Added hst_cos_twozxtab_0262.rmap,  results in added terminals, rmap
    Added COS TRACETAB as N/A
    Replaced COS XTRACTAB with OMIT
    Changed two COS FLATFILE references to OMIT and N/A

    >>> old_state = test_config.setup()
    >>> DiffScript("crds.diff crds://hst.pmap data/hst.pmap --recurse-added-deleted")()  # doctest: +ELLIPSIS
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'data/hst_cos_twozxtab_0262.rmap', ('FUV', 'SPECTROSCOPIC', '3.0'), ('2009-05-11', '00:00:00'), 'added terminal z2d1925ql_2zx.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_fluxtab.rmap', ('FUV', 'SPECTROSCOPIC'), ('1996-10-01', '00:00:00'), 'deleted terminal s7g1700kl_phot.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_fluxtab.rmap', ('FUV', 'SPECTROSCOPIC'), ('2009-05-11', '00:00:00'), 'deleted terminal u8k1433ql_phot.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_fluxtab.rmap', ('FUV', 'SPECTROSCOPIC'), ('2009-05-11', '00:00:01'), 'deleted terminal x1v17416l_phot.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_fluxtab.rmap', ('FUV', 'SPECTROSCOPIC'), ('2009-05-11', '00:00:02'), 'deleted terminal x6q17587l_phot.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_fluxtab.rmap', ('NUV', 'SPECTROSCOPIC'), ('1996-10-01', '00:00:00'), 'deleted terminal s7g17011l_phot.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_fluxtab.rmap', ('NUV', 'SPECTROSCOPIC'), ('2009-05-11', '00:00:00'), 'deleted terminal t9h1220sl_phot.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_xtractab.rmap', ('FUV', 'SPECTROSCOPIC', '-1|1'), ('2009-05-11', '00:00:01'), 'deleted terminal x1v17414l_1dx.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_xtractab.rmap', ('FUV', 'SPECTROSCOPIC', '-2147483648'), ('1996-10-01', '00:00:00'), 'deleted terminal s7g17006l_1dx.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_xtractab.rmap', ('FUV', 'SPECTROSCOPIC', '-2147483648'), ('2006-10-01', '00:00:00'), 'deleted terminal s7g17007l_1dx.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_xtractab.rmap', ('FUV', 'SPECTROSCOPIC', '2'), ('2009-05-11', '00:00:02'), 'deleted terminal x6q17586l_1dx.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_xtractab.rmap', ('NUV', 'SPECTROSCOPIC', '-1|1'), ('2009-05-11', '00:00:00'), 'deleted terminal w5g1439sl_1dx.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_xtractab.rmap', ('NUV', 'SPECTROSCOPIC', '-2147483648'), ('1996-10-01', '00:00:00'), 'deleted terminal s7g1700nl_1dx.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap',), 'hst_cos_xtractab.rmap', ('NUV', 'SPECTROSCOPIC', '-2147483648'), ('2006-10-01', '00:00:00'), 'deleted terminal s7g1700ol_1dx.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('hst_cos_flatfile.rmap', 'data/hst_cos_flatfile.rmap'), ('FUV', 'G130M|G140L|G160M'), ('1996-10-01', '00:00:00'), 'replaced n9n20182l_flat.fits with OMIT')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('hst_cos_flatfile.rmap', 'data/hst_cos_flatfile.rmap'), ('FUV', 'G160M'), ('1996-10-01', '00:00:00'), 'deleted Match rule for v4s17227l_flat.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('hst_cos_flatfile.rmap', 'data/hst_cos_flatfile.rmap'), ('NUV', 'G160M'), ('1996-10-01', '00:00:00'), 'added Match rule for v4s17227l_flat.fits')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('hst_cos_flatfile.rmap', 'data/hst_cos_flatfile.rmap'), ('NUV', 'G185M|G225M|G230L|G285M|MIRRORA|MIRRORB'), ('1996-10-01', '00:00:00'), 'replaced s7g1700tl_flat.fits with N/A')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('flatfile',), 'replaced hst_cos_flatfile.rmap with data/hst_cos_flatfile.rmap')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('fluxtab',), 'deleted hst_cos_fluxtab.rmap')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('tracetab',), 'added N/A')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('twozxtab',), 'added data/hst_cos_twozxtab_0262.rmap')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('hst_cos.imap', 'data/hst_cos.imap'), ('xtractab',), 'replaced hst_cos_xtractab.rmap with OMIT')
    (('.../mappings/hst/hst.pmap', 'data/hst.pmap'), ('cos',), 'replaced hst_cos.imap with data/hst_cos.imap')
    1
    >>> test_config.cleanup(old_state)
    """

def main():
    """Run module tests,  for now just doctests only.

    test_config.setup() and cleanup() are done inline above because bracketing
    the tests here does not get picked up by nose test discovery.  Combining
    tests into one giant docstring works but is hard to analyze and debug when
    things go wrong.
    """
    from crds.tests import test_diff, tstmod
    return tstmod(test_diff)

if __name__ == "__main__":
    print(main())



import os

from crds.core import rmap
from crds import config
from crds.hst import locate


_COS_PARAMETRIZED_COMPONENTS = {
    "cosmcp_g130mc1055m", "cosmcp_g130mc1096m", "cosmcp_g130mc1222m",
    "cosmcp_g130mc1291m", "cosmcp_g130mc1300m", "cosmcp_g130mc1309m",
    "cosmcp_g130mc1318m", "cosmcp_g130mc1327m", "cosmcp_g140lc1230m",
    "cosmcp_g140lc1280m", "cosmcp_g160mc1577m", "cosmcp_g160mc1589m",
    "cosmcp_g140lc1105m", "cosmcp_g140lc800m", "cosmcp_g160mc1533m",
    "cosmcp_g160mc1600m", "cosmcp_g160mc1611m", "cosmcp_g160mc1623m",
    "cosncm3g185mc1786m", "cosncm3g185mc1817m", "cosncm3g185mc1835m",
    "cosncm3g185mc1850m", "cosncm3g185mc1864m", "cosncm3g185mc1882m",
    "cosncm3g185mc1890m", "cosncm3g185mc1900m", "cosncm3g185mc1913m",
    "cosncm3g185mc1921m", "cosncm3g185mc1941m", "cosncm3g185mc1953m",
    "cosncm3g185mc1971m", "cosncm3g185mc1986m", "cosncm3g185mc2010m",
    "cosncm3g225mc2186m", "cosncm3g225mc2217m", "cosncm3g225mc2233m",
    "cosncm3g225mc2250m", "cosncm3g225mc2268m", "cosncm3g225mc2283m",
    "cosncm3g225mc2306m", "cosncm3g225mc2325m", "cosncm3g225mc2339m",
    "cosncm3g225mc2357m", "cosncm3g225mc2373m", "cosncm3g225mc2390m",
    "cosncm3g225mc2410m", "cosncm3g230lc2635m", "cosncm3g230lc2950m",
    "cosncm3g230lc3000m", "cosncm3g230lc3360m", "cosncm3g230lc3360m",
    "cosncm3g285mc2617m", "cosncm3g285mc2637m", "cosncm3g285mc2657m",
    "cosncm3g285mc2676m", "cosncm3g285mc2695m", "cosncm3g285mc2709m",
    "cosncm3g285mc2719m", "cosncm3g285mc2739m", "cosncm3g285mc2850m",
    "cosncm3g285mc2952m", "cosncm3g285mc2979m", "cosncm3g285mc2996m",
    "cosncm3g285mc3018m", "cosncm3g285mc3035m", "cosncm3g285mc3057m",
    "cosncm3g285mc3074m", "cosncm3g285mc3094m",
}

_ACS_PARAMETRIZED_COMPONENTS = {
    "acs_fr388n", "acs_fr388n", "acs_fr423n",
    "acs_fr459m", "acs_fr462n", "acs_fr505n",
    "acs_fr551n", "acs_fr601n", "acs_fr647m",
    "acs_fr656n", "acs_fr716n", "acs_fr782n",
    "acs_fr853n", "acs_fr914m", "acs_fr1016n",
    "acs_fr931n",
}

_WFPC2_PARAMETRIZED_COMPONENTS = {
    "wfpc2_contpc1", "wfpc2_contwf2", "wfpc2_contwf3",
    "wfpc2_contwf4",
}

_MISC_PARAMETRIZED_COMPONENTS = {
    "wfpc_contpc": "pc_cont#",
    "wfpc_contwf": "wf_cont#",
    "wfpc2_lrf": "wave#",
}

_GRAPH_KEYWORD_OVERRIDES = {
    # This oddball keyword is technically a bug in the graph, but
    # code in stsynphot and seemingly also IRAF exists to compensate
    # for it.
    "wfpc2_lrf": "lrf#",
    # Here "cont#" is intended as a sort of alias for "pc_cont#" or
    # "wf_cont#", so the graph will have a different value from
    # the lookup table or the throughput file.
    "wfpc_contwf": "cont#",
    "wfpc_contpc": "cont#",
}

# wfpc2 must occur before wfpc in this list, so that prefix
# checks will assign the correct instrument:
_HST_INSTRUMENTS = [
    "acs", "cos", "stis", "wfc3", "wfpc2",
    "foc", "hrs", "hsp", "nicmos", "wfpc",
    "ota", "fgs", "fos"
]

_NICMOS_COMPONENT_PREFIXES = {
    "nic", "nic1", "nic2", "nic3"
}

_OTA_COMPONENTS = {
    "clear", "dark", "hst_ota"
}

_NON_HST_INSTRUMENT = "nonhst"

ALL_INSTRUMENTS = _HST_INSTRUMENTS + [_NON_HST_INSTRUMENT]

GRAPH_REFTYPE = "tmgtab"
THROUGHPUT_REFTYPE = "thruput"
THROUGHPUT_LOOKUP_REFTYPE = "tmctab"
THROUGHPUT_FILENAME_SUFFIX = "syn"
THERMAL_REFTYPE = "thermal"
THERMAL_LOOKUP_REFTYPE = "tmttab"
THERMAL_FILENAME_SUFFIX = "th"
OBSMODES_REFTYPE = "obsmodes"
SYNPHOT_INSTRUMENT = "synphot"

# These are thermal components that are present in the lookup
# table but missing from the graph and individual component file
# headers. We can get rid of this once
# https://jira.stsci.edu/browse/REDCAT-100 is resolved.
KNOWN_MISSING_GRAPH_THERMAL_COMPONENTS = {
    "wfc3_ir_cor",
    "wfc3_ir_dn",
}

# These are thermal components that are present in the graph
# table but missing from the thermal lookup and individual component
# file headers.  We can get rid of this once
# https://jira.stsci.edu/browse/REDCAT-104 is resolved.
KNOWN_MISSING_THERMAL_LOOKUP_COMPONENTS = {
    "clear",
    "dark",
    "nic1_dn",
    "nic2_dn",
    "nic3_dn",
}

# These are throughput components that are present in the graph
# and lookup tables but missing from the individual file headers.
# We can get rid of this once
# https://jira.stsci.edu/browse/REDCAT-102 is resolved.
KNOWN_MISSING_THROUGHPUT_COMPONENTS = {
    "nic1_primary",
    "nic1_sec",
    "nic2_primary",
    "nic2_sec",
    "nic3_primary",
    "nic3_sec",
}

# These are thermal components that are present in the graph
# and lookup tables but missing from the individual file headers.
# We can get rid of this once
# https://jira.stsci.edu/browse/REDCAT-103 is resolved.
KNOWN_MISSING_THERMAL_COMPONENTS = {
    "nic1_bend", "nic1_bend1", "nic1_cmask", "nic1_dewar", "nic1_edge",
    "nic1_hole", "nic1_image", "nic1_pads", "nic1_para1", "nic1_para2",
    "nic1_primary", "nic1_pupil", "nic1_reimag", "nic1_sec", "nic1_spider",
    "nic2_sec", "nic3_bend", "nic3_bend1", "nic3_cmask", "nic3_dewar",
    "nic3_edge", "nic3_hole", "nic3_image", "nic3_pads", "nic3_para1",
    "nic3_para2", "nic3_primary", "nic3_pupil", "nic3_reimag", "nic3_sec",
    "nic3_spider"
}


def make_filename(reftype):
    """
    Create a unique name for the specified synphot graph or lookup reftype.
    """
    return locate.generate_unique_name_core(SYNPHOT_INSTRUMENT, reftype, ".fits")


def get_cache_path(context, reftype, error_on_missing=True):
    """
    Fetch the full path to the single reference file of reftype.
    """
    context = rmap.asmapping(context)
    imap = context.get_imap(SYNPHOT_INSTRUMENT)
    mapping = imap.get_rmap(reftype)
    filenames = mapping.reference_names()

    if len(filenames) == 0 and not error_on_missing:
        return None

    if len(filenames) != 1:
        raise RuntimeError("Expected '{}' rmap to contain one reference file".format(reftype))

    filename = filenames[0]
    return config.locate_file(filename, observatory="hst")


def get_instrument(component):
    """
    Get the *synphot* instrument corresponding to component.  Note that this
    is a superset of the CRDS HST instruments.
    """
    component = component.lower()

    for instrument in _HST_INSTRUMENTS:
        if component.startswith(instrument):
            return instrument

    if any(component.startswith(p) for p in _NICMOS_COMPONENT_PREFIXES):
        return "nicmos"

    if component in _OTA_COMPONENTS:
        return "ota"

    return _NON_HST_INSTRUMENT


def get_path_prefix(component):
    """
    Get the IRAF path prefix for the specified component.
    """
    instrument = get_instrument(component)
    return "cr{}comp$".format(instrument)


def get_lookup_filename(component, filename):
    """
    Get the lookup table filename for the specified component, which includes
    the IRAF path prefix and (if applicable) parametrized keyword suffix.
    """
    lookup_filename = get_path_prefix(component) + filename
    param_keyword = get_parametrization_keyword(component)
    if param_keyword is not None:
        lookup_filename = lookup_filename + "[{}]".format(param_keyword)
    return lookup_filename


def lookup_filename_to_path(synphot_root, lookup_filename):
    """
    Translate a filename from a lookup table into a full
    filesystem path.
    """
    prefix, other = lookup_filename.split("$")
    filename = other.split("[")[0]
    return os.path.join(synphot_root, "comp", prefix[2:-4], filename)


def get_mtab_path(synphot_root):
    """
    Get the full path to the synphot mtab directory.
    """
    return os.path.join(synphot_root, "mtab")


def get_instrument_path(synphot_root, instrument):
    """
    Get the full path to the synphot instrument's component directory.
    """
    return os.path.join(synphot_root, "comp", instrument)


def get_parametrization_keyword(component, is_graph=False):
    """
    Get the keyword prefix for a parametrized synphot component,
    or None if the component is not parametrized.
    """
    # Some of the individual files have uppercase COMPNAME, and there
    # are too many of them to be worth fixing.
    component = component.lower()

    if is_graph and component in _GRAPH_KEYWORD_OVERRIDES:
        return _GRAPH_KEYWORD_OVERRIDES[component]

    if component.endswith("_mjd") or component.endswith("_m"):
        return "mjd#"

    if component.endswith("_aper"):
        return "aper#"

    if component in _ACS_PARAMETRIZED_COMPONENTS:
        return component.split("_")[-1] + "#"

    if component in _COS_PARAMETRIZED_COMPONENTS:
        return "mjd#"

    if component in _WFPC2_PARAMETRIZED_COMPONENTS:
        return "cont#"

    if component in _MISC_PARAMETRIZED_COMPONENTS:
        return _MISC_PARAMETRIZED_COMPONENTS[component]

    return None



"""This module contains doctests and unit tests which exercise refactoring
code,  modules used to automatically update mappings.

"""
import os
from pprint import pprint as pp

import crds
from crds.core import log, exceptions
from crds import data_file, diff
from crds.refactoring import refactor
from crds.refactoring.refactor import RefactorScript
from crds.tests import test_config

from nose.tools import assert_raises, assert_true

# ==================================================================================

def dt_refactor_add_files():
    """
    >>> log.set_test_mode()

    >>> RefactorScript("crds.refactor insert data/hst_cos_deadtab.rmap hst_cos_deadtab_insert.rmap data/s7g1700hl_dead.fits")()  # doctest: +ELLIPSIS
    CRDS - INFO -  Inserting s7g1700hl_dead.fits into 'hst_cos_deadtab.rmap'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  1 infos
    0

    >>> diff.DiffScript("crds.diff data/hst_cos_deadtab.rmap ./hst_cos_deadtab_insert.rmap").run()
    (('data/hst_cos_deadtab.rmap', './hst_cos_deadtab_insert.rmap'), ('FUV',), ('1997-10-01', '01:01:01'), 'added terminal s7g1700hl_dead.fits')
    1

    >>> pp(refactor.rmap_check_modifications("data/hst_cos_deadtab.rmap", "./hst_cos_deadtab_insert.rmap", "none", "data/s7g1700hl_dead.fits", expected=("add",)))
    True

    >>> refactor.rmap_check_modifications("data/hst_cos_deadtab.rmap", "./hst_cos_deadtab_insert.rmap", "data/s7g1700gl_dead.fits", "data/s7g1700hl_dead.fits",
    ...                          expected=("replace",))
    CRDS - ERROR - Expected one of ('replace',) but got 'add' from change (('data/hst_cos_deadtab.rmap', './hst_cos_deadtab_insert.rmap'), ('FUV',), ('1997-10-01', '01:01:01'), "added terminal 's7g1700hl_dead.fits'")
    False

    >>> _ = os.system("rm ./*.rmap")

    """

def dt_refactor_delete_files():
    """
    >>> log.set_test_mode()

    >>> RefactorScript("refactor.py delete data/hst_cos_deadtab.rmap hst_cos_deadtab_delete.rmap data/s7g1700gl_dead.fits")()  # doctest: +ELLIPSIS
    CRDS - INFO -  Deleting 'data/s7g1700gl_dead.fits' from 'hst_cos_deadtab.rmap'
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  1 infos
    0

    >>> diff.DiffScript("diff.py data/hst_cos_deadtab.rmap ./hst_cos_deadtab_delete.rmap").run()
    (('data/hst_cos_deadtab.rmap', './hst_cos_deadtab_delete.rmap'), ('FUV',), ('1996-10-01', '00:00:00'), 'deleted Match rule for s7g1700gl_dead.fits')
    1

    >>> refactor.rmap_check_modifications("data/hst_cos_deadtab.rmap", "./hst_cos_deadtab_delete.rmap", "none", "data/s7g1700gl_dead.fits", expected=("delete_rule",))
    True

    >>> RefactorScript("refactor.py delete data/hst_cos_deadtab.rmap hst_cos_deadtab_delete2.rmap data/foobar.fits")()  # doctest: +ELLIPSIS
    CRDS - INFO -  Deleting 'data/foobar.fits' from 'hst_cos_deadtab.rmap'
    CRDS - ERROR -  Refactoring operation FAILED : Terminal 'foobar.fits' could not be found and deleted.
    CRDS - INFO -  1 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  1 infos
    1

    >>> os.path.exists("./hst_cos_deadtab_delete2.rmap")
    False

    """

def dt_refactor_add_header():
    """
    >>> log.set_test_mode()

    >>> RefactorScript("refactor.py set_header data/hst_cos_deadtab.rmap ./hst_cos_deadtab_add_header.rmap new_key some new value")()  # doctest: +ELLIPSIS
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  0 infos
    0

    >>> diff.DiffScript("diff.py data/hst_cos_deadtab.rmap ./hst_cos_deadtab_add_header.rmap --include-header-diffs --hide-boring-diffs").run()
    (('data/hst_cos_deadtab.rmap', './hst_cos_deadtab_add_header.rmap'), "header added 'new_key' = 'some new value'")
    1

    >>> pp(refactor.rmap_check_modifications("data/hst_cos_deadtab.rmap", "./hst_cos_deadtab_add_header.rmap", "none", "none", expected=("add_header",)))
    True

    >>> _ = os.system("rm ./*.rmap")
    """

def dt_refactor_replace_header():
    """
    >>> log.set_test_mode()

    >>> RefactorScript("refactor.py set_header data/hst_cos_deadtab.rmap ./hst_cos_deadtab_replace_header.rmap reffile_format something new")()  # doctest: +ELLIPSIS
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  0 infos
    0

    >>> diff.DiffScript("diff.py data/hst_cos_deadtab.rmap ./hst_cos_deadtab_replace_header.rmap --include-header-diffs --hide-boring-diffs").run()
    (('data/hst_cos_deadtab.rmap', './hst_cos_deadtab_replace_header.rmap'), "header replaced 'reffile_format' = 'table' with 'something new'")
    1

    >>> pp(refactor.rmap_check_modifications("data/hst_cos_deadtab.rmap", "./hst_cos_deadtab_replace_header.rmap", "none", "none", expected=("replace_header",)))
    True

    >>> _ = os.system("rm ./*.rmap")
    """

def dt_refactor_del_header():
    """
    >>> log.set_test_mode()

    >>> RefactorScript("refactor.py del_header data/hst_cos_deadtab.rmap ./hst_cos_deadtab_del_header.rmap reffile_format")()  # doctest: +ELLIPSIS
    CRDS - INFO -  0 errors
    CRDS - INFO -  0 warnings
    CRDS - INFO -  0 infos
    0

    >>> diff.DiffScript("diff.py data/hst_cos_deadtab.rmap ./hst_cos_deadtab_del_header.rmap --include-header-diffs --hide-boring-diffs").run()
    (('data/hst_cos_deadtab.rmap', './hst_cos_deadtab_del_header.rmap'), "deleted header 'reffile_format' = 'table'")
    1

    >>> pp(refactor.rmap_check_modifications("data/hst_cos_deadtab.rmap", "./hst_cos_deadtab_del_header.rmap", "none", "none", expected=("del_header",)))
    True

    >>> _ = os.system("rm ./*.rmap")
    """

def dt_refactor_bad_modify_count():
    """
    >>> log.set_test_mode()

    >>> refactor.rmap_check_modifications("data/hst_cos_deadtab.rmap", "data/hst_cos_deadtab_9998.rmap",
    ...                                   "data/s7g1700gl_dead.fits", "data/s7g1700hl_dead.fits", expected=("add",))
    CRDS - ERROR - Expected one of ('add',) but got 'replace' from change (('data/hst_cos_deadtab.rmap', 'data/hst_cos_deadtab_9998.rmap'), ('FUV',), ('1996-10-01', '00:00:00'), "replaced 's7g1700gl_dead.fits' with 's7g1700hl_dead.fits'")
    CRDS - ERROR - Expected one of ('add',) but got 'replace' from change (('data/hst_cos_deadtab.rmap', 'data/hst_cos_deadtab_9998.rmap'), ('NUV',), ('1996-10-01', '00:00:00'), "replaced 's7g1700ql_dead.fits' with 's7g1700hl_dead.fits'")
    False

    >>> refactor.rmap_check_modifications("data/hst_cos_deadtab.rmap", "data/hst_cos_deadtab_9998.rmap",
    ...                                   "data/s7g1700gl_dead.fits", "data/s7g1700hl_dead.fits", expected=("replace",))
    CRDS - ERROR - Replacement COUNT DIFFERENCE replacing 'data/s7g1700gl_dead.fits' with 'data/s7g1700hl_dead.fits' in 'data/hst_cos_deadtab.rmap' 1 vs. 2
    False
    """



# ==================================================================================

class TestRefactor(test_config.CRDSTestCase):

    '''
    example unit test code,  not for test_refactor

    def test_get_imap_except(self):
        r = rmap.get_cached_mapping("hst.pmap")
        with self.assertRaises(exceptions.CrdsUnknownInstrumentError):
            r.get_imap("foo")

    '''

# ==================================================================================


def tst():
    """Run module tests,  for now just doctests only."""

    import unittest
    suite = unittest.TestLoader().loadTestsFromTestCase(TestRefactor)
    unittest.TextTestRunner().run(suite)

    from crds.tests import test_refactor, tstmod
    return tstmod(test_refactor)

if __name__ == "__main__":
    print(tst())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# CRDS documentation build configuration file, created by
# sphinx-quickstart on Mon Nov 27 10:18:59 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import datetime
import importlib
import sys
import os
import sphinx
from distutils.version import LooseVersion

import sphinx_rtd_theme
# import sphinx_astropy
# from sphinx_astropy.conf import *

sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../../../crds'))

import crds

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    # 'numfig',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx_automodapi.automodapi',
    'sphinx_automodapi.automodsumm',
    'sphinx_automodapi.autodoc_enhancements',
    'sphinx_automodapi.smart_resolver',
    'sphinx_rtd_theme',
    ]


on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if on_rtd:
    extensions.append('sphinx.ext.mathjax')

elif LooseVersion(sphinx.__version__) < LooseVersion('1.4'):
    extensions.append('sphinx.ext.pngmath')
else:
    extensions.append('sphinx.ext.imgmath')

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'CRDS'
copyright = f'{datetime.datetime.now().year}, Space Telescope Science Institute'
author = 'Space Telescope Science Institute'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ".".join(crds.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags.
release = crds.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------
html_theme = "sphinx_rtd_theme"
html_theme_options = {
    'canonical_url': '',
    # 'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'vcs_pageview_mode': '',
    'style_nav_header_background': 'white',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False
}

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# import astropy_sphinx_theme
# html_theme_path = astropy_sphinx_theme.get_html_theme_path()
# html_theme = 'bootstrap-astropy'

# html_theme_options = {
#     'logotext1': 'crds',  # white,  semi-bold
#     'logotext2': '',  # orange, light
#     'logotext3': '',   # white,  light
#     'astropy_project_menubar': False
#     }

# html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
#        'donate.html',
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'CRDSdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, f'CRDS_users_guide-{release}.tex', 'CRDS Documentation',
     'Space Telescope Science Institute', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'crds', 'CRDS Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, f'CRDS_users_guide-{release}', 'CRDS Documentation',
     author, 'CRDS_users_guide', 'One line description of project.',
     'Miscellaneous'),
]



# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'obj'

# Don't show summaries of the members in each class along with the
# class' docstring
numpydoc_show_class_members = False

autosummary_generate = True

automodapi_toctreedirnm = 'api'

# Class documentation should contain *both* the class docstring and
# the __init__ docstring
autoclass_content = "both"

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    '-Nfontsize=10',
    '-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Efontsize=10',
    '-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Gfontsize=10',
    '-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif'
]


# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = False    DEPRECATED

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['globaltoc.html', 'relations.html', 'searchbox.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
# htmlhelp_basename = 'crdsdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '14pt',
    # Additional stuff for the LaTeX preamble.
    'preamble': r'''\usepackage{enumitem} \setlistdepth{99}'''
}

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = '_static/JWSTlogocrop.png'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
latex_show_urls = 'True'

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
latex_domain_indices = True

# Configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('http://matplotlib.org/', None),
    # 'jwst': ('https://jwst-pipeline.readthedocs.io/en/latest/jwst/package_index.html', None),
    }



"""This module defines helper functions that are used in "expression" validators,
either in a "presence" condition or as the "constraint" condition.   The expressions
themselves appear in .tpn files and are dynamically eval'ed in the context of a reference
header and other globals like these functions.

These functions are used to keep the .tpn expressions comparatively simple since those
are restricted to pigeon-Python that does not allow spaces.  See the JWST .tpn files
(particularly *array*.tpn) for examples of presence or constraint expressions,  grep
those files for these functions.
"""
from crds.core import utils, exceptions

# ----------------------------------------------------------------------------

# Table check expression helper functions

def has_columns(array_info, col_names):
    """Return True IFF CRDS `array_info` object defines `col_names` columns in any order.

    >>> has_columns(utils.Struct(COLUMN_NAMES=["THIS","THAT","ANOTHER"]), ["THAT","ANOTHER","THIS"])
    True
    >>> has_columns(utils.Struct(COLUMN_NAMES=["THIS","THAT","ANOTHER"]), ["THAT","ANOTHER","FOO"])
    False

    >>> has_columns("UNDEFINED",  ["THAT","ANOTHER","FOO"])
    False

    NOTE:  does not disallow extra columns not listed.
    """
    if not array_exists(array_info):
        return False
    for col in col_names:
        if col not in array_info.COLUMN_NAMES:
            return False
    return True

def has_type(array_info, typestr):
    """Return True IFF CRDS `array_info` object has a data array of type `typestr`.

    >>> has_type(utils.Struct({"DATA_TYPE" : "int8"}), "INT")
    True
    >>> has_type(utils.Struct({"DATA_TYPE" : "uint64"}), "INTEGER")
    True
    >>> has_type(utils.Struct({"DATA_TYPE" : "float64"}), "INTEGER")
    False
    >>> has_type(utils.Struct({"DATA_TYPE" : "float32"}), "FLOAT")
    True
    >>> has_type(utils.Struct({"DATA_TYPE" : "float32"}), "INT")
    False
    >>> has_type(utils.Struct({"DATA_TYPE" : "float64"}), "FLOAT")
    True
    >>> has_type(utils.Struct({"DATA_TYPE" : "complex32"}), "COMPLEX")
    True
    >>> has_type(utils.Struct({"DATA_TYPE" : "complex64"}), "COMPLEX")
    True
    >>> has_type(utils.Struct({"DATA_TYPE" : "complex64"}), "FLOAT")
    False
    >>> has_type(utils.Struct({"DATA_TYPE" : "complex64"}), ["FLOAT","INT"])
    False
    >>> has_type(utils.Struct({"DATA_TYPE" : "complex64"}), ["COMPLEX","INT"])
    True
    """
    possible_types = [typestr] if isinstance(typestr, str) else typestr
    for possible_type in possible_types:
        itype = _image_type(possible_type)
        if array_exists(array_info) and itype in array_info.DATA_TYPE:
            return True
    return False

def _image_type(typestr):
    """Return the translation of CRDS fuzzy type name `typestr` into numpy dtype str() prefixes.
    If CRDS has no definition for `typestr`,  return it unchanged.
    """
    return {
        'COMPLEX':'complex',
        'INT' : 'int',
        'INTEGER' : 'int',
        'FLOAT' : 'float',
        'BOOLEAN' : 'bool'
    }.get(typestr, typestr)

def has_column_type(array_info, col_name, typestr):
    """Return True IFF column `col_name` of CRDS `array_info` object has a
    data array of type `typestr`.

    >>> has_column_type(utils.Struct(DATA_TYPE={"WAVELENGTH":">f4"}), "WAVELENGTH", "FLOAT")
    True

    >>> has_column_type(utils.Struct(DATA_TYPE={"WAVELENGTH":">f4"}), "WAVELENGTH", "INTEGER")
    False

    >>> has_column_type("UNDEFINED", "WAVELENGTH", "INTEGER")
    False

    >>> has_column_type(utils.Struct(DATA_TYPE={"WAVELENGTH":">f4"}), "WAVELEN", "INTEGER")
    Traceback (most recent call last):
    ...
    crds.core.exceptions.MissingColumnError: Data type not defined for column 'WAVELEN'

    """
    if not array_exists(array_info):
        return False
    typestrs = _table_type(typestr)
    try:
        for typestr in typestrs:
            if typestr in array_info.DATA_TYPE[col_name.upper()]:
                return True
        return False
    except KeyError:
        pass
    raise exceptions.MissingColumnError("Data type not defined for column", repr(col_name))

def _table_type(typestr):
    """Return the translation of CRDS fuzzy type name `typestr` into numpy dtype str() prefixes.
    If CRDS has no definition for `typestr`,  return it unchanged.
    """
    int_types = [">i","<i","uint","int"]
    float_types = [">f","<f","float","float"]
    complex_types = [">c","<c","complex","complex"]
    string_types = ["|S"]

    def _array_types(types):
        return ["('" + typestr for typestr in types]

    trans = {
        'COMPLEX': complex_types,
        'COMPLEX_ARRAY': _array_types(complex_types),
        'INT' : int_types,
        'INTEGER' : int_types,
        'INT_ARRAY' : _array_types(int_types),
        'INTEGER_ARRAY' : _array_types(int_types),
        'FLOAT' : float_types,
        'FLOAT_ARRAY' : _array_types(float_types),
        'STR' : string_types,
        'STRING' : string_types,
        'STR_ARRAY' : _array_types(string_types),
        'STRING_ARRAY' : _array_types(string_types),
    }.get(typestr, typestr)

    return trans

def is_table(array_info):
    """Return True IFF CRDS `array_info` object corresponds to a table.

    >>> is_table(utils.Struct(KIND="IMAGE"))
    False
    >>> is_table(utils.Struct(KIND="TABLE"))
    True
    """
    return array_exists(array_info) and array_info.KIND=="TABLE"

def is_image(array_info):
    """Return True IFF CRDS `array_info` object corresponds to an image.

    >>> is_image(utils.Struct(KIND="IMAGE"))
    True
    >>> is_image(utils.Struct(KIND="TABLE"))
    False
    """
    return array_exists(array_info) and array_info.KIND=="IMAGE"

def  array_exists(array_info):
    """Return True IFF array_info is not UNDEFINED.

    >>> array_exists("UNDEFINED")
    False
    >>> array_exists(utils.Struct({"KIND":"IMAGE", "SHAPE" : (2048,2048), "TYPE": "float32"}))
    True
    """
    return array_info != "UNDEFINED"

def is_imaging_mode(exp_type):
    """Return True IFF `exp_type` is one of the imaging modes for any instrument.

    >>> is_imaging_mode('MIR_IMAGE')
    True
    >>> is_imaging_mode("NRS_IFU")
    False
    >>> is_imaging_mode('NRS_VERIFY')
    True
    """
    return exp_type in ["NRC_IMAGE", "NRC_TACQ", "NRC_TACONF", "NRC_CORON", "NRC_TSIMAGE",
                        "NRC_FOCUS", "NRC_DARK", "NRC_FLAT", "NRC_LED",

                        "MIR_IMAGE", "MIR_TACQ", "MIR_LYOT", "MIR_4QPM", "MIR_DARK",
                        "MIR_FLATIMAGE", "MIR_CORONCAL",

                        "NRS_CONFIRM", "NRS_DARK", "NRS_FOCUS", "NRS_IMAGE", "NRS_MIMF",
                        "NRS_TACONFIRM", "NRS_TACQ", "NRS_TASLIT", "NRS_VERIFY",

                        "NIS_IMAGE", "NIS_TACQ", "NIS_TACONFIRM", "NIS_AMI",
                        "NIS_FOCUS", "NIS_DARK", "NIS_LAMP",

                        "FGS_IMAGE", "FGS_FOCUS", "FGS_SKYFLAT", "FGS_INTFLAT", "FGS_DARK", "FGS_ID-STACK"]

def is_full_frame(subarray):
    """Return True IFF `subarray` is defined and has a full frame subarray value.

    >>> is_full_frame("UNDEFINED")
    False
    >>> is_full_frame("FULL")
    True
    >>> is_full_frame("GENERIC")
    True
    >>> is_full_frame("N/A")
    True
    >>> is_full_frame("ANY")
    True
    >>> is_full_frame("*")
    True
    >>> is_full_frame("MASK1140")
    False
    >>> is_full_frame("BRIGHTSKY")
    False
    """
    return subarray in ["FULL","GENERIC","N/A","ANY","*"]

def is_subarray(subarray):
    """Return True IFF `subarray` is defined and is not a full frame value.

    >>> is_subarray("UNDEFINED")
    False
    >>> is_subarray("GENERIC")
    False
    >>> is_subarray("N/A")
    False
    >>> is_subarray("ANY")
    False
    >>> is_subarray("*")
    False
    >>> is_subarray("FULL")
    False
    >>> is_subarray("MASK1140")
    True
    >>> is_subarray("BRIGHTSKY")
    True
    """
    return  (subarray != "UNDEFINED") and not is_full_frame(subarray)

def subarray_defined(header):
    """Return True IFF SUBARRAY related keywords are defined.

    >>> header = dict(SUBARRAY="GENERIC",SUBSTRT1="1",SUBSTRT2="1",SUBSIZE1="2048",SUBSIZE2="2048")
    >>> subarray_defined(header)
    True

    >>> header = dict(SUBARRAY="GENERIC",SUBSTRT1="1",SUBSTRT2="1",SUBISIZE2="2048")
    >>> subarray_defined(header)
    False
    """
    for keyword in ["SUBARRAY","SUBSTRT1","SUBSTRT2","SUBSIZE1","SUBSIZE2"]:
        value = header.get(keyword,"UNDEFINED")
        if value == "UNDEFINED":
            return False
    return True

def is_irs2(readpatt):
    """Return True IFF `readpatt` is one of the IRS2 READPATTs.

    >>> is_irs2("NRSIRS2")
    True
    >>> is_irs2("NRSIRS2RAPID")
    True
    >>> is_irs2("NRSN32R8")
    False
    >>> is_irs2("ALLIRS2")
    True
    """
    return 'IRS2' in readpatt

def is_defined(value):
    """Return True IFF `value` is not 'UNDEFINED' or None.

    >>> is_defined("UNDEFINED")
    False

    >>> is_defined(None)
    False

    >>> is_defined("FOO")
    True
    """
    return value not in ["UNDEFINED", None]

# # @utils.traced
# def nir_filter(instrument, reftype, exp_type):
#     """Return True if a SCI, ERR, or DQ array is appropriate for the specified
#     JWST NIR instrument, reftype, and exp_type.   This can be used to filter
#     out NIR SCI,ERR,DQ array definitions for those NIRSPEC modes and types
#     that don't define them.  The logic is too complex to inline.
#
#     >>> nir_filter("NIRSPEC", "SFLAT", "NRS_MSASPEC")
#     True
#     >>> nir_filter("NIRSPEC", "SFLAT", "NRS_IFU")
#     True
#     >>> nir_filter("NIRSPEC", "SFLAT", "NRS_FIXEDSLIT")
#     False
#     >>> nir_filter("NIRSPEC", "SFLAT", "NRS_BRIGHTOBJ")
#     False
#
#     >>> nir_filter("NIRSPEC", "DFLAT", "ANY")
#     True
#
#     >>> nir_filter("NIRSPEC", "FFLAT", "NRS_MSASPEC")
#     True
#     >>> nir_filter("NIRSPEC", "FFLAT", "NRS_IFU")
#     False
#     >>> nir_filter("NIRSPEC", "FFLAT", "NRS_FIXEDSLIT")
#     False
#     >>> nir_filter("NIRSPEC", "FFLAT", "NRS_BRIGTOBJ")
#     False
#
#     """
#     assert instrument != "MIRI",  "nir_filter() .tpn function should only be called for NIR-detector based instruments."
#     if instrument == "NIRSPEC":
#         if reftype == "SFLAT":
#             return exp_type in ["NRS_MSASPEC","NRS_IFU"]
#         elif reftype == "DFLAT":
#             return True
#         elif reftype == "FFLAT":
#             return exp_type in ["NRS_MSASPEC"]
#         elif reftype == "AREA":
#             return is_imaging_mode(exp_type)
#         else:
#             return True
#     else:
#         return True

# ----------------------------------------------------------------------------

# These are presence field helpers that mutate a True value of a presence expression
# into one of the non-boolean presence flags.   This enables combining presence values
# like "O" with presence expressions.   In general,  when an expression is True then
# the presence character of the wrapper helper is returned,  e.g.:
#
# (OPTIONAL(True)) --> "O"
# (OPTIONAL(False)) --> False,  constraint not evaluated.
#
# The enables having constraints which are conditionally optional,  so e.g. a constraint
# on

def optional(flag=True):
    """When this flag is True, an exception should be issued if the related keyword/element is
    defined and the constraint fails.   If the keyword/element is not defined or `flag` is False,
    the constraint should be ignored.   Returns "O" or False

    >>> optional(True)
    'O'
    >>> optional(False)
    False
    """
    return "O" if flag else False

def required(flag=True):
    """When this flag is True,  an exception should be issued if the related keyword/element is
    not defined.     Returns "R" or False.

    >>> required(True)
    'R'
    >>> required(False)
    False
    """
    return "R" if flag else False

def warning(flag=True):
    """Presence condition mutator/wrapper:

    When flag is True,  a warning should be issued if the related keyword/element is
    not defined.     Returns "W" or False.

    >>> warning(True)
    'W'
    >>> warning(False)
    False
    """
    return "W" if flag else False

def warn_only(flag):
    """Expression constraint mutator/wrapper:

    When flag is True,  the wrapped expression was satisifed, so return True
    signaling a passed validator expression.

    If the flag is False, the expression evaluated successfully but was not
    satisified.  Return the value "W" signaling that only a warning should be
    issued rather than an exception or error.

    >>> warn_only(True)
    True

    >>> warn_only(False)
    'W'
    """
    return "W" if not flag else True

def subarray(flag=True):
    """When this flag is True,  the related constraint should be applied if
    is_full_frame(SUBARRAY) is False.     Returns "S" or False.

    >>> subarray(True)
    'S'
    >>> subarray(False)
    False
    """
    return "S" if flag else False

def full_frame(flag=True):
    """When this flag is True,  the related constraint should be applied if
    is_full_frame(SUBARRAY) is True.     Returns "F" or False.

    >>> full_frame(True)
    'F'
    >>> full_frame(False)
    False
    """
    return "F" if flag else False

def all_subarray(flag=True):
    """When `flag` is True,  mark this constraint as applying to all SUBARRAY forms,
    including full frame, as long as SUBARRAY is defined.   Returns "A" or False.

    >>> all_subarray(True)
    'A'
    >>> all_subarray(False)
    False
    """
    return "A" if flag else False

# ----------------------------------------------------------------------------

def ndim(array, dims):
    """Return True IFF CRDS `array` object has number of dimensions `dims`.

    >>> array = utils.Struct({"KIND":"IMAGE", "SHAPE" : (2048,2048), "TYPE": "float32"})

    >>> ndim(array, 2)
    True

    >>> ndim(array, 3)
    False

    """
    return len(array.SHAPE) == dims

# ----------------------------------------------------------------------------

def test():
    import doctest
    from crds.certify import validator_helpers
    return doctest.testmod(validator_helpers)

if __name__ == "__main__":
    print(test())
